import threading,time,collections,sys,datetime,itertools,copy
from multiprocessing import Process,Manager,Queue
from . import archive,config
from .util import float_to_int,is_greater_list

#import Image
#import ImageFile
#import ImageFilter

class ImageConverter:
    def pixels_to_pilobject(self,size,pixels,stride,has_alpha=False,orientation=1):
        mode = 'RGBA' if has_alpha else 'RGB'
        return Image.frombuffer(mode,size,pixels,'raw',
                                mode,stride,orientation)

    def to_image_object(self,buffer):
        p = ImageFile.Parser()
        while True:
            dat = buffer.read(1024)
            if not dat:
                break
            p.feed(dat)
        img = p.close()
        return img

    def sharpen_image(self,pilobj):
        return pilobj.filter(ImageFilter.SHARPEN)

    def get_aspect_ratio(self,size):
        def get_common_factor(a,b):
            if b == 0:
                return a
            return get_common_factor(b,a % b)
        width,height = size
        result = get_common_factor(width,height)
        return [width / result,
                height /result]
    def resize(self,img_obj,new_size):
        return img_obj.resize(new_size,
                          Image.ANTIALIAS)
    def create_thumbnail(self,img,size):
        img_thumbnail = img.copy()
        img_thumbnail.thumbnail(size,Image.ANTIALIAS)
        return img_thumbnail

class ArchiveInfo:
    def __init__(self,archive_name=''):
        self.archive_name = archive_name
        self.current_page = 0
        self.total_page = 0

    def __str__(self):
        return '%s [ %d / %d ]' % (self.archive_name,
                                   self.current_page,
                                   self.total_page)

class ImageObjectFactory:
    def __init__(self,img_class_type):
        self.img_class_type = img_class_type
    def build(self,data):
        return self.img_class_type(data)

class AbstractImageObject(object):

    @property
    def size(self):pass

    @property
    def content(self):pass

    def resize(self,size,rotate_angle=None,need_flip=None,mode=None):pass

class AbstractImageFrame(object):

    @property
    def name(self):return ''
    
    def draw_image(self,img):pass

    def hide(self):pass

    def show(self):pass

class ImageDrawer:
    FIT_AUTO = 0
    FIT_MANUALLY = 1
    FIT_TO_WIDTH = 2
    FIT_TO_HEIGHT = 3

    __cur_img = None ; __fwd_img = None
    __box_size =0,0
    __container_size = 0,0
    __frame_content_inf ={}
    __mag_cache ={}
    __need_to_create_mag_cache = False
    __need_to_keep_rotation = False
    __flip = None
    __rotate_angle = None
    __is_switching_viewmode = False

    __is_dual = config.App.is_dualscreen_activate
    __fit_mode = config.App.image_fit_mode
    __is_right_to_left = config.App.is_right_to_left
    __magnifier_zoom_rate = config.App.magnifier_zoom_rate

    def __init__(self,img_frameR,img_frameL):
        self.conv = ImageConverter()
        self.img_frameR = img_frameR
        self.img_frameL = img_frameL
        if self.__is_dual:
            self.set_dualscreen_mode(self.__is_dual)

    def set_container_size(self,width,height):
        self.__container_size = width,height
        self.__update_box_size()

    def __update_box_alignment(self):
        minimum=0.0 ; center=0.5 ; maximum=1.0 
        frameL_align = center,center
        frameR_align = center,center
        if self.__is_dual and not self.__is_switching_viewmode:
            if self.__is_right_to_left:
                frameL_align = maximum,center
                frameR_align = minimum,center
            else:
                frameL_align = maximum,center
                frameR_align = minimum,center

        for frame,align in [[self.img_frameL,frameL_align],
                            [self.img_frameR,frameR_align]]:
            frame.set_alignment(align[0],align[1])

    def __update_box_size(self):
        self.__box_size = self.__container_size
        if self.__is_dual and not self.__is_switching_viewmode:
            self.__box_size = self.__box_size[0] / 2,self.__box_size[1]

    def redraw(self):
        frameR = self.img_frameR ; frameL = self.img_frameL
        cur = self.__cur_img ; fwd = self.__fwd_img

        if self.__is_dual:
            img_set = [frameR,cur],[frameL,fwd]
            if not self.__is_right_to_left:
                img_set = [frameL,cur],[frameR,fwd]
            
            if config.App.need_to_switch_singleview_once:
                self.__is_switching_viewmode = False
                for frame,img in img_set:
                    if img.size[0] > img.size[1]:
                        self.__is_switching_viewmode = True

                    if self.__is_switching_viewmode and img is not cur:
                        frame.hide()
                    else:
                        frame.show()

                self.__update_box_size()
                self.__update_box_alignment()

            for frame,img in img_set:
                self.__draw_frame(frame, img)
        else:
            self.__draw_frame(frameR, cur)
            self.__update_frame_inf(frameL,fwd)
    
    def draw_image(self,img,fwd=None):
        if not self.__need_to_keep_rotation:
            self.__clear_rotation()
        self.__cur_img = img ; self.__fwd_img = fwd
        self.redraw()

    def __update_frame_inf(self,frame,img):
        if img:
            zoom_rate, fit_size = self.__get_fit_size(frame,img)
            self.__frame_content_inf[frame.name] =(zoom_rate,fit_size)

    def __draw_frame(self,frame,img):
        fit_img = None
        if img:
            self.__update_frame_inf(frame,img)
            fit_size = self.__frame_content_inf[ frame.name ][1]
            if self.__need_to_create_mag_cache:
                self.__mag_cache[frame.name] = img.resize( 
                    tuple(
                        float_to_int(x * self.__magnifier_zoom_rate) for x in fit_size
                        ),
                    self.__rotate_angle,
                    self.__flip
                    )
            fit_img = img.resize( fit_size,
                                  self.__rotate_angle,
                                  self.__flip )

        frame.draw_image(fit_img)

    def __get_fit_size(self,frame,img):
        _zoom_rate = None
        img_info = self.__frame_content_inf.get(frame.name,None)
        if img_info:_zoom_rate = img_info[0]

        fit_size = None
        box_size = self.__box_size ; fit_mode = self.__fit_mode
        img_size = [p for p in img.size]

        if not config.App.need_to_expand_small_image:
            if is_greater_list(box_size,img_size):
                return 1.0,img_size

        if self.__rotate_angle:
            if self.__rotate_angle == 90 or self.__rotate_angle == 270:
                pass
                img_size.reverse()

        ratio = self.conv.get_aspect_ratio(img_size)

        def get_height_mode_size():
            con_h = box_size[1]
            factor = complex( con_h) / ratio[1]
            resized_width = float_to_int(ratio[0] * factor)
            zoom_rate = abs(complex(resized_width) / img_size[0])
            return zoom_rate,[resized_width,con_h]
        
        def get_width_mode_size():
            con_w = box_size[0]
            factor = complex(con_w) / ratio[0]
            resized_height = float_to_int(ratio[1] * factor)
            zoom_rate = abs(complex(resized_height) / img_size[1])
            return zoom_rate,[con_w,resized_height]

        if fit_mode == self.FIT_AUTO:
            con_w,con_h = box_size
            hzoom_rate,hmode = get_height_mode_size()
            wzoom_rate,wmode = get_width_mode_size()

            if is_greater_list(wmode,hmode):
                fit_size = hmode
                _zoom_rate = hzoom_rate
            else:
                fit_size = wmode
                _zoom_rate = wzoom_rate

        elif fit_mode == self.FIT_TO_HEIGHT:
            zoom_rate,fit_size = get_height_mode_size()
            _zoom_rate = zoom_rate

        elif fit_mode == self.FIT_TO_WIDTH:
            zoom_rate,fit_size = get_width_mode_size()
            _zoom_rate = zoom_rate

        elif fit_mode == self.FIT_MANUALLY:
            fit_size = [ float_to_int(val * _zoom_rate) for val in img_size  ]
 
        return _zoom_rate,fit_size

    def toggle_draw_direction(self,is_right_to_left):
        self.__is_right_to_left = is_right_to_left
        self.__update_box_alignment()

    def __clear_rotation(self):
        self.__flip = None
        self.__rotate_angle = None

    def set_fit_mode(self,fit_mode):
        if (not fit_mode == self.FIT_AUTO and
            not fit_mode == self.FIT_MANUALLY and
            not fit_mode == self.FIT_TO_WIDTH and
            not fit_mode == self.FIT_TO_HEIGHT):
            raise 'not defined fit mode'
        self.__fit_mode = fit_mode
        self.__draw_cache()

    def get_fit_mode(self):
        return self.__fit_mode

    def get_right_to_left_mode(self):
        return self.__is_right_to_left

    def set_right_to_left_mode(self,is_right_to_left):
        self.__is_right_to_left = is_right_to_left
        self.__draw_cache()

    def set_manual_zoom_rate(self,zoom_rates):
        infs = self.__frame_content_inf
        for zoom_rate in zoom_rates:
            for key in sorted(infs.keys()):
                infs[key] = (zoom_rate,
                             infs[key][1])

    def get_manual_zoom_rate(self):
        zoom_rates =[]
        infs = self.__frame_content_inf
        for key in sorted(infs.keys()):
            zoom_rates.append(infs[key][0])

        return zoom_rates

    @property
    def frame_content_info(self):
        return self.__frame_content_inf

    def get_magnifier_base(self,frame_name):
        return self.__mag_cache.get(frame_name,None)

    def need_to_create_mag_cache(self,need):
        self.__need_to_create_mag_cache = need
        if need:
            self.__draw_cache()
        else:
            self.__mag_cache.clear()

    def need_to_keep_rotation(self,is_need):
        if not is_need:
            self.__clear_rotation()

        self.__need_to_keep_rotation = is_need
        self.__draw_cache()

    def flip_image(self,is_horizontal):
        if not self.__flip:
            self.__flip = [False,False]
        
        if is_horizontal:
            self.__flip[0] = False if self.__flip[0] else True
        else:
            self.__flip[1] = False if self.__flip[1] else True

        self.__draw_cache()

    def rotate_image(self,angle):
        cur_angle = self.__rotate_angle
        if cur_angle is None:
            cur_angle = 0

        cur_angle = cur_angle + angle

        if cur_angle > 360:
            cur_angle = cur_angle - 360

        self.__rotate_angle = cur_angle
        self.__draw_cache()

    def get_dualscreen_mode(self):
        return self.__is_dual and not self.__is_switching_viewmode

    def set_dualscreen_mode(self,is_dual):
        self.__is_dual = is_dual
        if is_dual:
            self.img_frameL.show()
        else:
            self.img_frameL.hide()

        self.__update_box_size()
        self.__update_box_alignment()
        self.__draw_cache()

    def __draw_cache(self):
        if self.__cur_img or self.__fwd_img:
            self.redraw()

class ArchiveViewer:
    __namelist = None
    __cur_index = 0
    __archiver = None
    __terminate_signal =False
    __info = None
    def __init__(self,drawer,img_object_factory):
        self.drawer = drawer
        self.img_factory = img_object_factory

    def terminate(self):
        self.cache.terminate()
        if self.__archiver:
            self.__archiver.close()

    def run(self,storage,start_page_number=0,end_page_number=None):
        self.__info = ArchiveInfo()
        self.__init_archiver(storage)
        self.__init_viewer(start_page_number,end_page_number)
        self.cache.start()

        if isinstance(self.__archiver,archive.UnCompressedArchiveFile):
            for i in range( len( self.__namelist )):
                if storage.selected_path.endswith(self.__namelist[i]):
                    self.__cur_index = i
                    break
            self.__info.archive_name = storage.cwd_name
        else:
            self.__info.archive_name = storage.selected_file_name

        self.move_page_to(self.__cur_index)

    def __init_archiver(self,storage):
        self.__storage = storage
        self.__archiver = archive.get_archiver(storage)
        if not self.__archiver:
            raise 'not supported archiver'
            return

    def __init_viewer(self,start_page_number,end_page_number):
        self.__start_page = start_page_number
        self.__cur_index = start_page_number
        self.__namelist = self.__archiver.namelist()

        self.__total_page = end_page_number if end_page_number else len(self.__namelist)
        self.__info.current_page = start_page_number
        self.__info.total_page = self.__total_page
        self.cache = ImageCacheManager(self.__get_img,
                                       self.__total_page)
        
    def get_page_number(self):
        return self.__cur_index

    def get_page_sum(self):
        return self.__total_page

    def move_first_page(self):
        self.move_page_to(self.__start_page)

    def move_last_page(self):
        self.move_page_to(self.__total_page - 1)

    def move_next_page(self,degree=1):
        next_page = self.__cur_index + self.__get_page_padd() * degree
        if self.__total_page> next_page:
            self.move_page_to( next_page )
        else:
            self.move_page_to( self.__total_page -1 )

    def move_prev_page(self,degree=1):
        prev_page = self.__cur_index - self.__get_page_padd() * degree
        if self.__start_page <= prev_page:
            self.move_page_to( prev_page )
        else:
            self.move_page_to( self.__start_page )

    def move_page_to(self,num):
        if num > self.__total_page -1 or num < self.__start_page:
            raise 'cannot move to invalid page number' 
        self.__cur_index = num
        self.__info.current_page = num

        cur = self.__get_img_with_cache(num)

        fwd = None

        if num + 1 < self.__total_page:
            fwd = self.__get_img_with_cache(num + 1)

        self.drawer.draw_image( cur,fwd )

    def __get_img_with_cache(self,num):
        img = self.cache.get(num)
        if not img:
            while not self.__terminate_signal and not img:
                img = self.cache.get(num)
                time.sleep(0.0001)
        return self.img_factory.build( img )

    def __get_page_padd(self):
        if not self.drawer.get_dualscreen_mode():
            padd = 1
        else:
            padd = 2
        return padd
    '''
    def __regist_prev_and_next_key(self,num):
        total_csize = config.App.page_cache_size
        prev_csize = total_csize / 2
        next_csize = total_csize - prev_csize
        processed_index = None

        def pickup_updated_keys(standerd):
            keyset = self.cache.keyset()
            nkey_end = standerd + next_csize
            if nkey_end > self.__total_page:nkey_end = self.__total_page

            pkey_start = standerd - prev_csize
            if pkey_start <0:pkey_start = 0

            pkey_range = range(pkey_start,standerd)
            nkey_range = range(standerd,nkey_end)
            
            #collect out-range keys to cache out them
            delkeys = keyset.difference( pkey_range + nkey_range)
            
            #collect new keys to cache in them
            p_newkeys = sorted(  set(pkey_range).difference(keyset) , reverse=True)
            n_newkeys = sorted(  set(nkey_range).difference(keyset) )
            return delkeys, p_newkeys, n_newkeys

        delkeys,prevs,nexts = pickup_updated_keys(num)
        self.cache.remove(delkeys)
        self.cache.regist(nexts)
        self.cache.regist(prevs)
        '''
    def __get_img(self,num):
        return self.__archiver.read( self.__namelist[num]) 

    @property
    def info(self):
        return self.__info

class _RequestKeyBuilder:
    def __init__(self,total_page):
        total_csize = config.App.page_cache_size
        self.prev_csize = total_csize / 2
        self.next_csize = total_csize - self.prev_csize
        self.__total_page = total_page

    def build(self,standerd,keyset):
        nkey_end = standerd + self.next_csize
        if nkey_end > self.__total_page:nkey_end = self.__total_page

        pkey_start = standerd - self.prev_csize
        if pkey_start <0:pkey_start = 0

        pkey_range = range(pkey_start,standerd)
        nkey_range = range(standerd,nkey_end)
            
            #collect out-range keys to cache out them
        delkeys = keyset.difference( pkey_range + nkey_range)

            #collect new keys to cache in them
        p_newkeys = sorted(  set(pkey_range).difference(keyset) , reverse=True)
        n_newkeys = sorted(  set(nkey_range).difference(keyset) )
        yield delkeys
        yield p_newkeys
        yield n_newkeys

class ImageCacheManager:
    def __init__(self,get_img_func,total_page):
        obj_proxy = Manager()
        self.__cache = obj_proxy.dict()
        self.shared_proxy = obj_proxy
        self.request = Queue()
        self.__worker = Process(target=self.extracter,
                                args=(self.request,
                                      self.__cache,
                                      total_page,
                                      get_img_func,)
                                )
    def start(self):
        self.__worker.start()

    def terminate(self):
        self.__worker.terminate()

    def get(self,page):
        if page not in  self.__cache:
            self.request.put(page)
        return self.__cache.get(page,None)
            
    def extracter(self,reqq,cache,total_page,get_img_func):
        pendding = collections.deque()
        request = _RequestKeyBuilder(total_page)
        def remove_key(dusts):
            for page in dusts:
                if page in cache:del cache[page]
                elif page in pendding:pendding.remove(page)

        def recv_request(standard):
            registed_keyset = set(cache.keys()).union( pendding )
            keyset_iter = request.build( standard,
                                         registed_keyset)
            remove_key( keyset_iter.next() )
            for entrys in keyset_iter:
                for entry in entrys:pendding.appendleft(entry)

        while True:
            if not reqq.empty():
                recv_request( reqq.get() )

            if len(pendding) > 0:
                page = pendding.pop()
                cache[page] = get_img_func(page)
            else:
                time.sleep(0.01)
'''
class _SharedObjectManager(SyncManager):pass
class _CacheKeyManager(collections.deque):
    __poped_key = None
    
    @property
    def poped_key(self):
        return self.__poped_key

    def empty(self):
        return len(self) == 0

    def append_end(self,x):
        print x
        self.appendleft(x)

    def pop(self):
        key = self.pop()
        print 'pop!:' + str(key)
        self.__poped_key = key
        return key

class ImageCacheManager:

    __reqq = _CacheKeyManager()
    __cache = {}
    __worker = None
    __terminate_signal = False
    __processing_page = None

    def __init__(self,get_img_func):
        self.__get_img = get_img_func
        obj_proxy = _SharedObjectManager()
        obj_proxy.start()
        self.__cache = obj_proxy.dict()
        self.shared_proxy = obj_proxy

    def regist(self,pages):
#        print 'reg:'+str(pages)
        for page in pages:
            if (not page in self.__reqq and 
                not page in self.__cache and 
                page !=self.__reqq.poped_key):
                self.__reqq.append_end(page)

        if not self.__worker:
            self.__worker = Process(target=self.__async_get_img,
                                            args=(self.__reqq.pop,
                                                  self.__cache,
                                                  self.__get_img,)
                                            )
            self.__worker.start()

#        print self.__reqq

    def remove(self,pages):
        for page in pages:
            if page in self.__cache:
                del self.__cache[page]
            elif page in self.__reqq:
                self.__reqq.remove(page)
#        print 'rm:'+str(pages)

    def terminate(self):
        if self.__worker:
            self.__worker.terminate()
            self.__worker.join(0.5)
            self.shared_proxy.shutdown()

        self.__cache.clear()
        self.__reqq.clear()

    def keyset(self):
        req_and_cache_set = set(self.__cache.keys()).union( 
            self.__reqq
            )
        if self.__reqq.poped_key:
            req_and_cache_set = req_and_cache_set.union([self.__reqq.poped_key])
        return req_and_cache_set

    def get(self,page):
        return self.__cache.get(page,None)

    def __async_get_img(self,req_pop,cache,get_img):
        while True:
            
            page = None
            try:
                page = req_pop()
                print 'page:'+str(page)
                cache[page] = get_img(page)

            except IndexError:
                time.sleep(1)

            except:
                print sys.exc_info()
                break

class ImageCacheManager:
    __reqq = collections.deque()
    __cache = {}
    __active_thread = None
    __terminate_signal = False
    __processing_page = None

    def __init__(self,get_img_func):
        self.__get_img = get_img_func
        
    def regist(self,pages):
#        print 'reg:'+str(pages)
        if not self.__active_thread:
            self.__active_thread = threading.Thread(target=self.__worker)
            self.__active_thread.setDaemon(True)
            self.__active_thread.start()
        for page in pages:
            if (not page in self.__reqq and 
                not page in self.__cache and 
                page !=self.__processing_page):
                self.__reqq.appendleft(page)
#        print self.__reqq

    def remove(self,pages):
        for page in pages:
            if page in self.__cache:
                del self.__cache[page]
            elif page in self.__reqq:
                self.__reqq.remove(page)
#        print 'rm:'+str(pages)

    def terminate(self):
        if self.__active_thread:
            self.__terminate_signal = True
            self.__active_thread.join(0.5)

        self.__cache.clear()
        self.__reqq.clear()

    def keyset(self):
        req_and_cache_set = set(self.__cache.keys()).union( self.__reqq)
        if self.__processing_page:
            req_and_cache_set = req_and_cache_set.union([self.__processing_page])
        return req_and_cache_set

    def get(self,page):
        return self.__cache.get(page,None)

    def __worker(self):
        try:
            while not self.__terminate_signal:
                
                if len(self.__reqq)==0:
                    continue

                self.__processing_page = self.__reqq.pop()
                page = self.__processing_page
                self.__cache[page] = self.__get_img(page)
                print page
        except:
            print sys.exc_info()
'''
