# -*- coding: utf-8 -*-
import struct,datetime,inspect,itertools,subprocess

COMPRESSION_TYPE =['-lh0-','-lhd-'
                   '-lh4-','-lh5-','-lh6-']

OS_IDENTIFIER = {-1:'Unkown',
                  0x4d:'MS-DOS',
                 0x32:'IBM OS/2',
                 0x39:'Microware OS-9',
                 0x48:'OS/68K',
                 0x55:'UNIX',
                 0x43:'CP/M',
                 0x46:'TSC FLEX',
                 0x52:'Runser',
                 0x54:'TownsOS',
                 0x58:'XOSK',
                 0x77:'Windows95',
                 0x57:'WindowNT',
                 0x00:'MS-DOS',
                 0x4a:'Java'}

LEVELED_HEADER_MAP = {}
EXTENDED_HEADER_MAP = {}

class LhaFile(object):
    __factory = None
    def __init__(self,file):
        self.file = BinaryFile(file)
        self. headers = {}
        self.__load_headers()

    def __load_headers(self):

        generator = LhaHeaderGenerator(self.file)
        for header in generator:
            self.headers[header.filename] = header
            
    def close(self):
        self.file.close()

    def getinfo(self,name):
        header = self.headers.get(name,None)
        if header:
            header.parse_content_info(self.file)

        return header

    def infolist(self):
        for entry in self.headers.namelist():
            yield self.getinfo(entry)

    def namelist(self):
        return self.headers.keys()

    def read(self,name):
        target = self.headers.get(name,None)
        content = None
        if target:
            if name.endswith('/'):
                content = self.__read_dir(name)
            else:
                if not self.__factory:
                    self.__factory = LhaContentDecoderFactory()
                decoder = self.__factory.create(self.file,target)
                content = decoder.decode()
                del decoder

        return content

    def __read_dir(self,name):
        #pickup filename from archive-entry
        entries = itertools.ifilter(lambda key:
                                        key.startswith(name) and not key.endswith('/'),
                                    self.namelist()
                                    )

        for entry in entries:
            yield self.read(entry)

class MSDosDateFormat:
    def __init__(self,date,time):
        self.__parse_date(date)
        self.__parse_time(time)

    def __parse_date(self,date):
        date_l,date_h = [ord(val) for val in date]
        past_year = date_h >> 1
        self.year = 1980 + past_year
        h_month = date_h - (past_year << 1)
        l_month = date_l >> 5
        self.month = h_month + l_month 
        self.day = date_l - (l_month << 5)

    def __parse_time(self,time):
        time_l,time_h = [ord(val) for val in time]
        self.hour = time_h >> 3
        h_min = time_h - ( self.hour << 3)
        l_min = time_l >> 5
        self.min = (h_min << 3) + l_min
        self.sec = ( time_l - (l_min << 5) ) * 2

    def to_tuple(self):
        return (self.year,self.month,self.day,
                self.hour,self.min,self.sec)

class CommonHeader(object):
    @staticmethod
    def id():
        return 0x00

    def __init__(self,file,header_size):
        self.size = header_size
        self.position = file.tell()
        self.next_header_size = BinaryUtil.conv_to_int( 
            file.peek(2,header_size -2,1)
            )
    
    def parse(self,file):
        pass

class FileName(CommonHeader):

    @staticmethod
    def id():
        return 0x01
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

    def parse(self,file):
        rawdata =  file.peek(self.size -3,self.position + 1)
        self.name = rawdata

class DirectoryName(CommonHeader):

    @staticmethod
    def id():
        return 0x02
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

    def parse(self,file):
        rawdata =  file.peek(self.size -3,self.position + 1)
        self.name = rawdata.replace(chr(0xff),'/')

class MSDosAttribute(CommonHeader):

    @staticmethod
    def id():
        return 0x40
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class WindowsTimestamp(CommonHeader):

    @staticmethod
    def id():
        return 0x41
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class FileSize(CommonHeader):

    @staticmethod
    def id():
        return 0x01
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class UnixPermission(CommonHeader):

    @staticmethod
    def id():
        return 0x50
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class UnixOwnerId(CommonHeader):

    @staticmethod
    def id():
        return 0x51

    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class UnixGrantedGroup(CommonHeader):

    @staticmethod
    def id():
        return 0x52

    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class UnixGrantedUser(CommonHeader):

    @staticmethod
    def id():
        return 0x53

    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class UnixTimestamp(CommonHeader):

    @staticmethod
    def id():
        return 0x54

    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class MultiOsFileAttribute2(CommonHeader):

    @staticmethod
    def id():
        return 0x7F
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class MultiOsFileAttribute(CommonHeader):

    @staticmethod
    def id():
        return 0xFF
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class OS2FileAttribute(CommonHeader):

    @staticmethod
    def id():
        return 0x7E
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class Comment(CommonHeader):

    @staticmethod
    def id():
        return 0x3F
    
    def __init__(self,file,header_size):
        CommonHeader.__init__(self,file,header_size)

class ExtendedHeaderGenerator(object):
    def __init__(self,file,first_header_size):
        self.file = file
        self.__first_header_size = first_header_size
        self.header_map = EXTENDED_HEADER_MAP

    def __iter__(self):
        next_header_size = self.__first_header_size
        start_position = self.file.tell()
        exception = None
        while next_header_size:
            header_id = BinaryUtil.conv_to_int( self.file.peek(1))
            header_type = self.header_map.get(header_id,None)
            if header_type:
                header = header_type(self.file,next_header_size)
                self.file.seek(next_header_size,1)
                next_header_size = header.next_header_size
                yield header
            else:
                exception = TypeError,'undefined extended-header-id:%s parsed' % (hex(header_id))
        self.file.seek(start_position)
        if exception:
            raise exception

class UnixExtendedArea:
    def __init__(self,file):
        self.parse_datetime(file)

    def parse_owner_info(self,file):
        self.parse_permission(file)
        self.parse_user_id(file)
        self.parse_group_id(file)

    def parse_datetime(file):
        self.date_time =None

    def parse_permission(self,file):
        pass

    def parse_user_id(self,file):
        pass

    def parse_group_id(self,file):
        pass

class LhaHeader(object):
    is_already_info_parsed = False
    def __init__(self,file):
        self.position = file.tell()
        self.parse_compress_type(file)
        if self.compress_type in COMPRESSION_TYPE:
            self.parse_header_level(file)
            self.parse_entry_size(file)
            self.is_broken = False
        else:
            self.is_broken = True

    def parse_content_info(self,file):
        if self.is_already_info_parsed:return
        file.seek(self.position)
        self.parse_file_size(file)
        self.parse_file_crc(file)
        self.parse_date_time(file)
        self.parse_create_system(file)
        self.is_already_info_parsed = True

    def parse_entry_size(self,file):
        self.parse_header_size(file)
        self.entry_size = self.header_size

    def parse_header_level(self,file):
        self.header_level = BinaryUtil.conv_to_int( file.peek(1,20,1) )

    def parse_compress_type(self,file):
        self.compress_type = file.peek(5,2,1)

    def parse_header_size(self,file):
        self.header_size = BinaryUtil.conv_to_int( file.peek(1,0,1) ) + 2

class LhaLevel0Header(LhaHeader):
    def __init__(self,file):
        LhaHeader.__init__(self,file)
        self.parse_filename(file)

    def parse_filename(self,file):
        self.__path_length = BinaryUtil.conv_to_int( file.peek(1,21,1) )
        self.filename = file.peek(self.__path_length,
                                  22,1)

    def parse_entry_size(self,file):
        LhaHeader.parse_entry_size(self,file)
        self.parse_compress_size(file)
        self.entry_size = self.entry_size + self.compress_size

    def parse_compress_size(self,file):
        self.compress_size = BinaryUtil.conv_to_int( file.peek(4,7,1) )

    def parse_date_time(self,file):
        date = file.peek(2,17,1)
        time = file.peek(2,15,1)
        self.date_time = MSDosDateFormat(date,time).to_tuple()

    def parse_file_size(self,file):
        self.file_size = BinaryUtil.conv_to_int( file.peek(4,11,1) )

    def parse_file_crc(self,file):
        self.CRC = BinaryUtil.conv_to_int(
            file.peek(2,22 + self.__path_length,1)
            )

    def parse_create_system(self,file):
        os_id = BinaryUtil.conv_to_int(file.peek(1,24 + self.__path_length,1))
        #if exten-field exists,overwrite os-id value as 'unix'
        if os_id == 0xFF:
            os_id == 0x55
        elif os_id not in OS_IDENTIFIER:
            os_id = -1
        self.create_system = os_id
        
    @staticmethod
    def id():
        return 0x00

class LhaLevel1Header(LhaLevel0Header):
    extended_header = {}
    def __init__(self,file):
        LhaLevel0Header.__init__(self,file)
#        self.header_size = self.header_size

    def parse_filename(self,file):
        LhaLevel0Header.parse_filename(self,file)
        older_define = self.filename if self.filename else ''
        self.parse_filename_with_ext_header(file)

        if self.filename:
            self.filename = self.filename + older_define
        else:
            self.filename = older_define

    def parse_filename_with_ext_header(self,file):
        self.parse_extended_header(file)
        self.filename =''
        for header_type in DirectoryName,FileName:
            header = self.extended_header.get(header_type.id(),None)

            if header:
                header.parse(file)
                if header.name:
                    self.filename = self.filename + header.name

    def parse_extended_header(self,file):
        next_header_size = BinaryUtil.conv_to_int(
            file.peek(2,self.header_size-2,1)
            )
        file.seek(self.header_size,1)

        extended_header_size = 0
        for header in ExtendedHeaderGenerator(file,next_header_size):
            extended_header_size = extended_header_size + header.size
            self.extended_header[header.id()] = header

        self.header_size = self.header_size + extended_header_size
        self.compress_size = self.compress_size - extended_header_size
        file.seek(self.position)

    @staticmethod
    def id():
        return 0x01

class LhaLevel2Header(LhaLevel1Header):
    extended_header = {}
    def __init__(self,file):
        LhaLevel0Header.__init__(self,file)

    def parse_filename(self,file):
        LhaLevel1Header.parse_filename_with_ext_header(self,file)

    def parse_header_size(self,file):
        self.header_size = BinaryUtil.conv_to_int( file.peek(2,0,1) )

    def parse_extended_header(self,file):
        file.seek(24,1)
        next_header_size = BinaryUtil.conv_to_int(
            file.peek(2,0,1)
            )
        file.seek(2,1)
        for header in ExtendedHeaderGenerator(file,next_header_size):
            self.extended_header[header.id()] = header
        file.seek(self.position)
        
    def parse_date_time(self,file):
        self.date_time = datetime.datetime.fromtimestamp(
            BinaryUtil.conv_to_int( file.peek(4,15,1) )
            ).timetuple()[0:6]

    def parse_file_crc(self,file):
        self.CRC = BinaryUtil.conv_to_int(
            file.peek(2,21,1)
            )

    def parse_create_system(self,file):
        self.create_system = BinaryUtil.conv_to_int(file.peek(1,23,1))

    @staticmethod
    def id():
        return 0x02
        
class LhaContentDecoderFactory:
    __decoder_suffix = 'Decoder'
    def __init__(self):
        self.__load_decoder()

    def __load_decoder(self):
        host_module = inspect.getmodule(self)
        decoders = itertools.ifilter(lambda attr: attr.endswith(self.__decoder_suffix),
                                     dir(host_module) )

        self.decoder_map = {}
        for decoder in decoders:
            self.decoder_map[decoder] = getattr(host_module,decoder)

    def __get_decoder(self,header):
        method_name = header.compress_type
        decoder_name = method_name.strip('-').title() + self.__decoder_suffix
        return self.decoder_map[decoder_name]

    def create(self,file,header):
        decoder = self.__get_decoder(header)(file,header)
        return decoder

class Lh0Decoder(object):
    def __init__(self,file,header):
        self.header = header
        self.file = file

    def decode(self):
        header = self.header
        return self.file.peek(header.compress_size,
                              header.position + header.header_size)

class LhXDecoder(object):
    def __init__(self,file,header):
        self.header = header
        self.file = file

    def decode(self):
        command = ['lha','pq',self.file.name,self.header.filename]
        fileobj = subprocess.Popen(command,
                                   stdout=subprocess.PIPE).communicate()[0]
        return fileobj

class Lh4Decoder(LhXDecoder):
    pass

class Lh5Decoder(LhXDecoder):
    pass

class Lh6Decoder(LhXDecoder):
    pass

class LhaHeaderGenerator(object):
    def __init__(self,file):
        self.file = file
        self.header_map = LEVELED_HEADER_MAP

    def __iter__(self):
        header = LhaHeader(self.file)
        position = self.file.tell()
        exception = None
        while header:
            if header.is_broken:
                if header.compress_type:
                    exception = (TypeError,
                                 'unavaliable compress-type:' + \
                                     header.compress_type + \
                                     ' parsed @ '+ \
                                     hex(header.position)
                                     )
                break

            header_level = header.header_level
            header = self.header_map.get(header_level,None)

            if header:
                header = header(self.file)
                self.file.seek(header.entry_size,1)
                yield header
                header = LhaHeader(self.file)
            else:
                exception = TypeError,'undefined level-header parsed'
                break

        self.file.seek(position)
        if exception:
            raise exception[0],exception[1]

class BinaryUtil(object):
    @staticmethod
    def conv_to_int(data,is_little_endian=True):
        int_data = None
        if len(data) > 3:
            endian = '<' if is_little_endian else '>'
            int_data = struct.unpack(endian + 'i', data )[0]
        else:
            little_endianed_data = data if is_little_endian else data[::-1]
            int_data = 0
            for i in range(len(data)):
                int_data = int_data + (
                    ord( little_endianed_data[i] ) << (8 *i)
                    )
        return int_data

class BinaryFile(object):
    def __init__(self,filename,mode='rb',bufsize=-1):
        self.fileobj = open(filename,mode,bufsize)
        self.__attach_fileobj_method()

    def __attach_fileobj_method(self):
        attr_iter = itertools.ifilter(lambda name:not name.startswith('__'),
                                      dir(self.fileobj))
        for attr in attr_iter:
            attrobj = getattr(self.fileobj,attr)
            if callable(attrobj):
                setattr(
                    self,attr,attrobj
                    )

    @property
    def closed(self):
        return self.fileobj.closed

    @property
    def name(self):
        return self.fileobj.name

    def peek(self,size,pos=0,whense=0):
        prev = self.fileobj.tell()
        if pos:
            self.fileobj.seek(pos,whense)
        data = self.fileobj.read(size)
        self.fileobj.seek(prev)
        return data

def __regist_header_map(map,header_class_list):
    map.update((header.id(),header) for header in header_class_list )

def __regist_leveled_header_map():
    __regist_header_map(LEVELED_HEADER_MAP,
                        (LhaLevel0Header,
                         LhaLevel1Header,
                         LhaLevel2Header)
                        )
def __regist_extended_header_map():
    __regist_header_map(EXTENDED_HEADER_MAP,
                        (CommonHeader,
                         Comment,
                         FileSize,
                         FileName,
                         DirectoryName,
                         WindowsTimestamp,
                         MSDosAttribute,
                         OS2FileAttribute,
                         MultiOsFileAttribute,
                         MultiOsFileAttribute2,
                         UnixPermission,
                         UnixOwnerId,
                         UnixTimestamp,
                         UnixGrantedGroup,
                         UnixGrantedUser)
                        )

__regist_leveled_header_map()
__regist_extended_header_map()

def print_info(lhaobj,entry):
    inf = lhaobj.getinfo(entry)
    for attr_name in dir(inf):
        if not attr_name.startswith('__'):
            attr = getattr(inf,attr_name)
            if not callable(attr):
                print attr_name + ':'+str(attr)        

def print_namelist(lhaobj):
    for entry in sorted(lhaobj.namelist()):
        print entry
